/********************************************************/
/* Copyright 1999, Chengjie Tu & Trac Duy Tran          */
/* All rights reserved                                  */
/* File: "huffman.C"                                    */
/* Description: Huffman codec class                     */
/* Affiliation: Johns Hopkins University, Baltimore, MD */
/*              FastVideo LLC, Columbia, MD             */
/********************************************************/

#include<iostream.h>
#include<iomanip.h>
#include"huffman.h"
#include"bitstream.h"

Huffman::Huffman(unsigned msize,int* ms,unsigned* mp)
  :_size(msize),_symbol(ms),_frequency(mp)
{
  _root=new Node(ms[0],mp[0],0,0,0);  //first node
  
  for(unsigned c=1;c<_size;c++)  //create symbol list in ascending order
  {
    Node* foo=new Node(ms[c],mp[c],0,0,0);  //new node
    
    if(_root->count<foo->count)  //new node can't be root
    {
      Node* bar;
      for(bar=_root;bar->next&&bar->next->count<foo->count;bar=bar->next); //where new node should be
      foo->next=bar->next, bar->next=foo;      //insert new node
    }
    else   //new node to be root
      foo->next=_root,_root=foo;
  }

  for(unsigned c=1;c<_size;c++)  //keep on merging two most unlikely symbols
  {                              //until only one node exists
    Node* bar;
    unsigned foo=_root->count+_root->next->count; //merge frequency
    for(bar=_root->next;bar->next&&bar->next->count<foo;bar=bar->next); //where new node should be
    Node* gah=new Node(0,foo,_root,_root->next,bar->next); //merge nodes, merged nodes as new node's children
    bar->next=gah;              //insert new node into list
    _root=_root->next->next;    //disconnect merged nodes from list
  }

#ifdef ENCODER
  _code_book=new BitList*[_size];  //initialize codebook
  for(unsigned c=0;c<_size;c++)
    _code_book[c]=0;
  _code_tree(_root);               //get codebook
#endif
}

Huffman::~Huffman(void)
{
  _delete_tree(_root);           //clear up codetree

#ifdef ENCODER  
  for(unsigned i=0;i<_size;i++)  //clean up codebook
    for(BitList* foo=_code_book[i];_code_book[i];) //delete bit list
    {
      foo=_code_book[i];
      _code_book[i]=_code_book[i]->next;
      delete foo;
    }
  delete _code_book;
#endif
}

#ifdef ENCODER
bool Huffman::CodeSymbol(int s,OutBitStream& out)
{
  unsigned i;

  for(i=0;i<_size&&s!=_symbol[i];i++); //find symbol index
  if(i==_size)                         //not a legal symbol
    return false;
  
  for(BitList* foo=_code_book[i];foo;foo=foo->next)
    out.CodeBit(foo->bit);            //code it
  
  return true;
}
#else
int Huffman::DecodeSymbol(InBitStream& in)
{
  Node* foo=_root;
  
  while(foo->left||foo->right) //travel from root to leaf
    foo=(in.DecodeBit()?foo->right:foo->left); //'0'/'1' to right/left child
  
  return foo->symbol;  //got it
}
#endif

void Huffman::_delete_tree(Node* root)  //delete codetree
{
  Node* left=root->left,*right=root->right;
  delete root; //delete this node
  if(left)     //delete left tree
    _delete_tree(left);
  if(right)    //delete right tree
    _delete_tree(right);
}

#ifdef ENCODER
void Huffman::_code_tree(Node* root) //update codebook
{
  if((!root->left)&&(!root->right))  //leaf tree node
  {
    unsigned i;

    for(i=0;root->symbol!=_symbol[i];i++); //find symbol index
    for(Node* foo=root;foo->next;foo=foo->next)  //travel from leaf to root
    {
      BitList* bar=new BitList(foo->count==1);   //left/right child --->'0'/'1'
      bar->next=_code_book[i];                   //insert to the HEAD of bitlist
      _code_book[i]=bar;
    }
    return;
  }
  
  if(root->left) //have left child
  {
    root->left->next=root;  //get parent
    root->left->count=0;    //mark it as left
    _code_tree(root->left); //recurse left tree
  }
  if(root->right) //have right child
  {
    root->right->next=root; //get parent
    root->right->count=1;   //mark it as right
    _code_tree(root->right); //recurse right tree
  }
}

void Huffman::ShowCodeBook(ostream& out)
{
  out<<"    symbol frequency  huffmancode"<<endl<<endl;
  for(unsigned i=0;i<_size;i++)
  {
    out<<setw(10)<<_symbol[i]<<setw(10)<<_frequency[i]<<": ";
    for(BitList* foo=_code_book[i];foo;foo=foo->next)
      out<<foo->bit;
    out<<endl;
  }
}
#endif
